/* eslint-env node */

import {
  access,
  copyFile,
  mkdir,
  readdir,
  stat,
  writeFile,
  readFile,
  unlink,
} from "fs/promises";
import { constants } from "fs";
import { join, basename, dirname } from "path";
import { zip } from "zip-a-folder";

// A vite build plugin to customize the output of vite builds for our extension

// based on https://github.com/goosy/vite-plugin-copy/blob/master/src/index.js
// keep the attribution until this is rewritten
async function prepareDir(dir) {
  const parents = dirname(dir);
  await access(parents, constants.F_OK).catch(async (err) => {
    if (err.code === "ENOENT") await prepareDir(parents);
  });
  await access(dir, constants.F_OK).catch(async (err) => {
    if (err.code === "ENOENT")
      await mkdir(dir).catch((err) => {
        if (err.code !== "EEXIST") console.log(err);
      });
  });
}

/**
 * exists returns a promise that resolves to an object containing
 * a boolean flag `exists` which is true if the target exists, and
 * a file stat object for that file for further use.
 *
 * @param target {string} path to the target directory or file to test
 * @return {Promise<{stat: fs.Stats, exists: boolean}>} a result with existence test data
 */
async function exists(target) {
  const result = {
    exists: true,
    stat: undefined,
  };
  result.stat = await stat(target).catch((err) => {
    if (err.code === "ENOENT") {
      result.exists = false;
    }
  });
  return result;
}

// copySubDir takes a `src` directory and copies its contents recursively
// into `dest`. This method differs from copyToFile because it does not
// copy the `src` directory itself to dest, only it's contents.
async function copySubDir(src, dest) {
  const srcExists = await exists(src);
  if (!srcExists.exists || !srcExists.stat.isDirectory()) {
    return;
  }

  await prepareDir(dest);

  for (const file of await readdir(src)) {
    const srcPath = join(src, file);
    await copyToDir(srcPath, dest);
  }
}

async function copyManifest(src, dest) {
  const dir = dirname(dest);
  await prepareDir(dir);

  const srcExists = await exists(src);
  if (!srcExists.exists || srcExists.stat.isDirectory()) {
    throw "could not find src " + src;
  }

  // locate actual background script
  let backgroundFile;
  for (const file of await readdir(join("dist", "assets", "worker"))) {
    if (file.startsWith("service.") && file.endsWith(".js")) {
      backgroundFile = file;
    }
  }

  // load manifest
  const manifestBuffer = await readFile(src);
  const manifest = manifestBuffer.toString();
  const finalManifest = manifest.replaceAll(
    "BACKGROUND_SCRIPT",
    `assets/worker/${backgroundFile}`
  );

  await writeFile(dest, Buffer.from(finalManifest));
}

async function rmSourceMaps(dir) {
  const maps = [];
  for (const file of await readdir(dir)) {
    const srcPath = join(dir, file);
    const srcStat = await stat(srcPath);
    if (srcStat.isDirectory()) {
      await rmSourceMaps(srcPath);
    } else {
      if (file.endsWith(".js.map")) {
        // source map - remove it
        maps.push(srcPath);
      }
    }
  }
  await Promise.all(maps.map((src) => unlink(src)));
}

async function copyToDir(src, dst) {
  await prepareDir(dst);

  const srcExists = await exists(src);

  if (!srcExists.exists) return;

  const dstPath = join(dst, basename(src));
  if (srcExists.stat.isDirectory()) {
    for (const file of await readdir(src)) {
      const srcPath = join(src, file);
      await copyToDir(srcPath, dstPath);
    }
  } else {
    await copyFile(src, dstPath)
      .then(() => { }) //console.log(`copied ${src} to ${dstPath}`))
      .catch((err) => {
        console.log(`The file "${src}" could not be copied ${err}`);
      });
  }
}

// bundles is an array of bundle subdirectory names that should have the
// `dist` directory copied into it.
const bundles = [
  { name: "edge", manifest: "edge", version: 3, sourcemaps: true },
  { name: "chrome", manifest: "chrome", version: 2, sourcemaps: true },
  { name: "firefox", manifest: "firefox", version: 2, sourcemaps: true },
];

// legacy bundles that use manifest v2
const bundlesLegacy = [];

/**
 * Add uBlock extension specific build steps as a Vite plugin.
 *
 * @param [legacy] {boolean} true if this is a legacy (manifest v2) build
 * @returns {{closeBundle: ((function(): Promise<void>)|*), buildEnd: ((function(): Promise<void>)|*), name: string}}
 */
export function uBlockBuild(legacy) {
  // noinspection JSUnusedGlobalSymbols
  return {
    name: "build-extension",
    closeBundle: async () => {
      let tasks = [];
      // Copy content scripts, dist, and manifests
      const b = legacy ? bundlesLegacy : bundles;
      b.forEach((bundle) => {
        tasks.push(copySubDir("./dist", `./bundles/${bundle.name}`));
        tasks.push(
          copyToDir(
            "./node_modules/@eyeo/webext-ad-filtering-solution/dist/ewe-content.js",
            `./bundles/${bundle.name}`
          )
        );
        tasks.push(
          copyManifest(
            `./manifest/${bundle.manifest}.json`,
            `./bundles/${bundle.name}/manifest.json`
          )
        );
        tasks.push(
          copyToDir(
            "./node_modules/webextension-polyfill/dist/browser-polyfill.js",
            `./bundles/${bundle.name}`
          )
        );
        tasks.push(
          copyToDir(
            "./node_modules/webextension-polyfill/dist/browser-polyfill.js.map",
            `./bundles/${bundle.name}`
          )
        );
      });
      tasks.push(
        copyToDir(
          "./node_modules/webextension-polyfill/dist/browser-polyfill.js",
          "./public"
        )
      );
      tasks.push(
        copyToDir(
          "./node_modules/webextension-polyfill/dist/browser-polyfill.js.map",
          "./public"
        )
      );
      await Promise.all(tasks);
      // Create final zips
      tasks = [];
      for (let i = 0; i < b.length; i++) {
        const bundle = b[i];
        const bundleDir = `./bundles/${bundle.name}`;

        if (bundle.sourcemaps) {
          // we leave source maps in place
        } else {
          // remove source maps
          await rmSourceMaps(bundleDir);
        }
        tasks.push(
          zip(`./bundles/${bundle.name}`, `./bundles/${bundle.name}.zip`)
        );
      }
      await Promise.all(tasks);
    },
    buildEnd: async () => { },
  };
}
